# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from pathlib import PurePosixPath
from urllib.parse import urljoin, urlparse, parse_qs
import dateutil.parser
import re

from .common import get_relative_url, normalize_url, regex_match
from .common import HtmlExtractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup, SoupTag

if TYPE_CHECKING:
    from requests import Response


class PhpbbExtractor(HtmlExtractor):
    tests = [
        {
            "url": "https://phpbb.com/community",
            "test_base_url": "https://www.phpbb.com/community/",
            "test_boards": {
                ("47",): {
                    "title": "General",
                },
                ("551",): {
                    "title": "Support Forums",
                },
                ("451",): {
                    "title": "Extensions Forums",
                },
                ("471",): {
                    "title": "Styles Forums",
                },
                ("52",): {
                    "title": "Non-support Specific",
                },
                ("48",): {
                    "title": "phpBB Archives",
                },
            },
        },
        {
            "url": "https://phpbb.com/community/viewforum.php?f=556",
            "item_count": 0,
            "test_base_url": "https://www.phpbb.com/community/",
            "test_boards": {
                ("551", "556", "561"): {
                    "title": "[3.2.x] Convertors",
                },
                ("551", "556", "566"): {
                    "title": "[3.2.x] Translations",
                },
            },
        },
        {
            "url": "https://www.phpbb.com/community/viewtopic.php?t=2377611",
            "test_base_url": "https://www.phpbb.com/community/",
            "test_contents_hash": "d98797411be739205db624165f24fc248c93157b",
            "test_item_count": 7,
        },
        {
            "url": "https://www.phpbb.com/community/viewtopic.php?t=2534156",
            "test_base_url": "https://www.phpbb.com/community/",
            "test_contents_hash": "d88e2da2e01e843f71c70f62a874542a416efc5d",
            "test_item_count": 1,
        },
        {
            "url": "https://wikipediocracy.com/forum",
            "test_base_url": "https://wikipediocracy.com/forum/",
            "test_boards": {
                ("11",): {
                    "title": "Public Area",
                },
                ("11", "13"): {
                    "title": "Mission Statement, Terms of Service, and Welcome",
                },
                ("11", "8"): {
                    "title": "Wikipedia and Wikimedia Projects - General Discussion",
                },
                ("11", "8", "14"): {
                    "title": "Governance",
                },
                ("11", "8", "16"): {
                    "title": "Biographies (BLPs) & Privacy",
                },
                ("11", "8", "23"): {
                    "title": "The Money Trail",
                },
                ("11", "8", "17"): {
                    "title": "Jimboland",
                },
                ("11", "8", "42"): {
                    "title": "Sexism",
                },
                ("11", "8", "43"): {
                    "title": "Technology",
                },
                ("11", "8", "25"): {
                    "title": "Blog Posts",
                },
                ("11", "6"): {
                    "title": "News and Media",
                },
                ("11", "21"): {"title": "Web 2.0: The Emperor's New Clothes"},
            },
        },
        {
            "url": "https://wikipediocracy.com/forum/viewtopic.php?f=13&t=7003",
            "test_base_url": "https://wikipediocracy.com/forum/",
            "test_contents_hash": "173a8a4e8558ec8eebe419ecc8274be406054275",
            "test_item_count": 115,
        },
        {
            "url": "https://forum.scssoft.com/",
            "test_base_url": "https://forum.scssoft.com/",
            "test_board_count": 250,
        },
        {
            "url": "https://forum.scssoft.com/viewforum.php?f=137",
            "test_base_url": "https://forum.scssoft.com/",
            "test_boards": {
                ("137", "139"): {
                    "title": "American Truck Simulator",
                },
                ("137", "249"): {
                    "title": "Euro Truck Simulator 2",
                },
                ("137", "152"): {
                    "title": "ETS2 - dashboards at night",
                },
            },
        },
        {
            "url": "https://forum.scssoft.com/viewtopic.php?t=245658",
            "test_base_url": "https://forum.scssoft.com/",
            "test_contents_hash": "5982a8dfb095eb30f1e318342ba63741786c92c6",
            "test_item_count": 21,
        },
        {
            "url": "https://forums.themanaworld.org",
            "test_base_url": "https://forums.themanaworld.org/",
            "test_board_count": 30,
        },
        {
            "url": "https://forums.themanaworld.org/viewforum.php?f=82",
            "test_base_url": "https://forums.themanaworld.org/",
            "test_boards": {
                ("27", "82", "11"): {
                    "title": "Web Development",
                },
                ("27", "82", "63"): {
                    "title": "Tutorials",
                },
            },
        },
        {
            "url": "https://forums.themanaworld.org/viewtopic.php?t=362",
            "test_base_url": "https://forums.themanaworld.org/",
            "test_item_count": 37,
            "test_contents_hash": "25065476c54fdc7a60ee934b630a6b975a4e98dc",
        },
        {
            "url": "https://forum.minetest.net/",
            "test_base_url": "https://forum.minetest.net/",
            "test_board_count": 48,
        },
        {
            "url": "https://forum.minetest.net/",
            "test_base_url": "https://forum.minetest.net/",
            "test_board_count": 48,
        },
        {
            "url": "https://forum.minetest.net/viewforum.php?f=48",
            "test_base_url": "https://forum.minetest.net/",
            "test_boards": {
                ("38", "48", "15"): {
                    "title": "Game Releases",
                },
                ("38", "48", "49"): {
                    "title": "Game Discussion",
                },
                ("38", "48", "50"): {
                    "title": "WIP Games",
                },
                ("38", "48", "51"): {
                    "title": "Old Games",
                },
            },
        },
        {
            "url": "https://forum.minetest.net/viewtopic.php?f=51&t=9066",
            "test_base_url": "https://forum.minetest.net/",
            "test_item_count": 76,
            "test_contents_hash": "f370438d250e2d9b36412b83c4a5c8391f2cc38f",
            # TODO: Test for embeds.
        },
    ]

    _board_item_css = "a.topictitle"
    _board_next_page_css = ".next a"
    _thread_item_css = "div.post"
    _thread_next_page_css = ".next a"

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        # Check for the existence of "viewforum.php".
        response = session.try_get(
            urljoin(
                normalize_url(url, remove_suffixes=["viewforum.php", "viewtopic.php"]),
                "viewforum.php",
            ),
            should_cache=True,
            should_retry=False,
        )

        # A rather crude way to detect: we look for a <html> tag with "dir" attribute.
        soup = Soup(response.text)
        soup.find("html", attrs={"dir": True})

        return PhpbbExtractor(
            session,
            normalize_url(
                response.url, remove_suffixes=["viewforum.php", "viewtopic.php"]
            ),
            options,
        )

    def _is_viewforum_url(self, url: str):
        parsed_url = urlparse(url)
        path = PurePosixPath(str(parsed_url.path))

        if len(path.parts) != 1:
            try:
                get_relative_url(url, self.base_url)
            except ValueError:
                return False

        if path.parts[-1] != "viewforum.php":
            return False

        parsed_query = parse_qs(parsed_url.query)

        if "f" not in parsed_query:
            return False

        return True

    def _is_viewtopic_url(self, url: str, thread_id: int | None = None):
        parsed_url = urlparse(url)
        path = PurePosixPath(str(parsed_url.path))

        if len(path.parts) != 1:
            try:
                get_relative_url(url, self.base_url)
            except ValueError:
                return False

        if path.parts[-1] != "viewtopic.php":
            return False

        parsed_query = parse_qs(parsed_url.query)

        if "t" not in parsed_query:
            return False

        return True

    def _is_viewtopic_pagination_url(self, url: str, thread_id: int | None = None):
        if not self._is_viewtopic_url(url, thread_id):
            return False

        parsed_url = urlparse(url)
        parsed_query = parse_qs(parsed_url.query)

        if not {"t", "start"} <= parsed_query.keys():
            return False

        return True

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url, should_cache=True)
        soup = Soup(response.content)

        board_lis = soup.find_all("div", class_="forabg")

        for board_li in board_lis:
            header_li = board_li.find("li", class_="header")
            board_id = None

            if board_anchor := header_li.try_find("a"):
                parsed_href = urlparse(board_anchor.get("href"))
                parsed_query = parse_qs(parsed_href.query)

                try:
                    board_id = parsed_query["f"][0]
                except KeyError:
                    continue

                board_title = board_anchor.string
                board_url = urljoin(self.base_url, f"viewforum.php?f={board_id}")

                self._set_board(
                    path=(board_id,),
                    url=board_url,
                    origin=response.url,
                    data={},
                    title=board_title,
                    are_subboards_fetched=True,
                )

            subboard_anchors = board_li.find_all("a", class_="forumtitle")

            for subboard_anchor in subboard_anchors:
                parsed_href = urlparse(subboard_anchor.get("href"))
                parsed_query = parse_qs(parsed_href.query)

                try:
                    subboard_id = parsed_query["f"][0]
                except KeyError:
                    continue

                subboard_title = subboard_anchor.string
                subboard_url = urljoin(self.base_url, f"viewforum.php?f={subboard_id}")

                self._set_board(
                    path=(board_id, subboard_id) if board_id else (subboard_id,),
                    url=subboard_url,
                    origin=response.url,
                    data={},
                    title=subboard_title,
                )

    def _do_fetch_subboards(self, board: Board):
        if board is self.root:
            return

        response = self._session.get(board.url, should_cache=True)

        try:
            get_relative_url(response.url, self.base_url)
        except ValueError:
            return

        soup = Soup(response.content)

        subboard_anchors = soup.find_all("a", class_="forumtitle")

        for subboard_anchor in subboard_anchors:
            parsed_href = urlparse(subboard_anchor.get("href"))
            parsed_query = parse_qs(parsed_href.query)

            try:
                subboard_id = parsed_query["f"][0]
            except KeyError:
                continue

            subboard_title = subboard_anchor.string
            subboard_url = urljoin(self.base_url, f"viewforum.php?f={subboard_id}")

            self._set_board(
                path=board.path + (subboard_id,),
                url=subboard_url,
                origin=response.url,
                data={},
                title=subboard_title,
                are_subboards_fetched=True,
            )

    def _resolve_url(self, url: str):
        return normalize_url(
            self._session.get(url, should_cache=True).url, keep_queries=["f", "t"]
        )

    def _get_node_from_url(self, url: str):
        response = self._session.get(url, should_cache=True)
        resolved_url = normalize_url(response.url, keep_queries=["f", "t"])

        parsed_url = urlparse(resolved_url)
        parts = PurePosixPath(parsed_url.path).parts

        if parts[-1] == "viewforum.php":
            self._fetch_lower_boards(self.root)

            parsed_query = parse_qs(parsed_url.query)

            if "f" not in parsed_query:
                return self.root

            board_id = parsed_query["f"][0]

            for board in self._boards:
                if board is not self.root and board.path[-1] == board_id:
                    return board

            raise ValueError
        elif parts[-1] == "viewtopic.php":
            topic_id = parse_qs(parsed_url.query)["t"][0]
            soup = Soup(response.content)
            breadcrumbs = soup.find(class_="breadcrumbs")

            breadcrumb_anchors = breadcrumbs.find_all("a", attrs={"itemprop": "item"})

            breadcrumb_urls = [
                self._resolve_url(urljoin(url, anchor.get("href")))
                for anchor in breadcrumb_anchors
            ]
            board = self.find_board_from_urls(tuple(breadcrumb_urls[1:]))

            title_h2 = soup.find("h2", class_="topic-title")
            title = title_h2.find("a").string

            return Thread(
                path=board.path + (topic_id,),
                url=resolved_url,
                origin=resolved_url,
                data={},
                title=title,
            )
        elif normalize_url(resolved_url) == self.base_url:
            return self.root

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, subboard_id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _extract_board_page_thread(
        self, board: Board, state: PageState, response: Response, tag: SoupTag
    ):
        href = urljoin(self.base_url, tag.get("href"))
        parsed_href = urlparse(href)
        parsed_query = parse_qs(parsed_href.query)
        thread_id = parsed_query["t"][0]

        return Thread(
            path=board.path + (thread_id,),
            url=href,
            origin=response.url,
            data={},
            title=tag.string,
        )

    def _extract_thread_page_post(
        self, thread: Thread, state: PageState, response: Response, tag: SoupTag
    ):
        id_div_regex = re.compile(r"^post_content(\d+)$")
        id_div = tag.find("div", id=id_div_regex)
        content_div = tag.find("div", class_="content")

        author_p = tag.find("p", class_="author")

        username_tag = author_p.find(
            {"a", "span"}, class_={"username", "username-coloured"}
        )

        if time_tag := author_p.try_find("time"):
            creation_time = time_tag.get("datetime")
        else:
            # Date-string begins right after &raquo;.
            date_match = re.search("»(.+)", author_p.tag.get_text(), re.MULTILINE)

            if date_match:
                creation_time = dateutil.parser.parse(date_match.group(1)).isoformat()
            else:
                raise ValueError

        url_h3 = tag.find("h3")
        url_anchor = url_h3.find("a")

        return Post(
            path=thread.path,
            subpath=(regex_match(id_div_regex, id_div.get("id")).group(1),),
            url=urljoin(response.url, url_anchor.get("href")),
            origin=response.url,
            data={},
            author=username_tag.string,
            creation_time=creation_time,
            content=str("".join(str(v) for v in content_div.contents)),
        )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from abc import ABC, abstractmethod
from pydantic import BaseModel
from urllib.parse import urljoin, urlparse, urlunparse, parse_qs, urlencode
from pathlib import PurePosixPath
import logging
import traceback

from ..session import Session
from ..soup import Soup, SoupTag
from ..exceptions import AttributeSearchError, SearchError
from ..version import __version__

if TYPE_CHECKING:
    from requests import Response


def get_relative_url(url: str, base_url: str):
    parsed_base_url = urlparse(base_url)
    parsed_url = urlparse(url)

    base_path = PurePosixPath(str(parsed_base_url.path))
    path = PurePosixPath(str(parsed_url.path))

    if str(base_path) == ".":
        return path

    return str(path.relative_to(base_path))


def normalize_url(
    url: str,
    remove_suffixes: list[str] = ["index.php"],
    append_slash: bool = True,
    keep_queries: list[str] = [],
):
    parsed_url = urlparse(url)
    new_path = parsed_url.path.removesuffix("/")

    if not keep_queries or not parsed_url.query:
        for remove_suffix in remove_suffixes:
            new_path = new_path.removesuffix(remove_suffix)

    new_path = new_path.removesuffix("/")

    query = parse_qs(parsed_url.query)
    new_query = {key: query[key] for key in keep_queries if key in query}

    new_parsed_url = parsed_url._replace(
        path=new_path, params="", query=urlencode(new_query, doseq=True), fragment=""
    )

    new_url = urlunparse(new_parsed_url)

    if append_slash and not new_parsed_url.query:
        return f"{new_url}/"

    return str(new_url)


def regex_match(pattern: Pattern[str], strings: list[str] | str):
    if isinstance(strings, str):
        strings = [strings]

    for string in strings:
        result = pattern.match(string)

        if result:
            return result

    raise ValueError


def regex_search(pattern: Pattern[str], strings: list[str] | str):
    if isinstance(strings, str):
        strings = [strings]

    for string in strings:
        result = pattern.search(string)

        if result:
            return result

    raise ValueError


class ExtractorOptions(BaseModel):
    path: bool


class PageState(BaseModel):
    url: str
    page: int


class Item(BaseModel):
    path: tuple[str, ...]
    url: str
    origin: str
    data: dict[str, Any]


class Post(Item):
    subpath: tuple[str, ...]
    author: str
    creation_time: str
    content: str


class Thread(Item):
    title: str


class Board(Item):
    title: str


class File(Item):
    subpath: tuple[str, ...]
    content_type: str | None = None
    content: bytes | None = None
    os_path: str | None = None


class Extractor(ABC):
    tests: list[dict[str, Any]]

    @final
    @classmethod
    def detect(
        cls, session: Session, url: str, options: ExtractorOptions
    ) -> Extractor | None:
        try:
            return cls._detect(session, url, options)
        except SearchError:
            pass

    @staticmethod
    @abstractmethod
    def _detect(
        session: Session, url: str, options: ExtractorOptions
    ) -> Extractor | None:
        pass

    def __init__(self, session: Session, base_url: str, options: ExtractorOptions):
        self._session = session
        self.base_url = base_url
        self.root = Board(
            path=(), url=self._resolve_url(base_url), origin=base_url, data={}, title=""
        )
        self._boards: list[Board] = [self.root]
        self._subboards: dict[tuple[str, ...], dict[str, Board]] = {(): {}}
        self._are_subboards_fetched: dict[tuple[str, ...], bool] = {(): False}
        self._are_all_boards_fetched: bool = False
        self._options = options

        self.board_state: PageState | None = None
        self.thread_state: PageState | None = None

    @final
    def fetch(self):
        self._fetch_top_boards()
        # self._fetch_lower_boards(self.root)

    @abstractmethod
    def _fetch_top_boards(self):
        pass

    def _set_board(
        self,
        *,
        path: tuple[str, ...],
        replace_path: tuple[str, ...] | None = None,
        are_subboards_fetched: bool | None = None,
        **kwargs: Any,
    ):
        if replace_path is None:
            replace_path = path

        parent_board = self._find_board(replace_path[:-1])

        if replace_path[-1] in self._subboards[parent_board.path]:
            for k, v in kwargs.items():
                setattr(self._subboards[parent_board.path][replace_path[-1]], k, v)

            new_parent_board = self._find_board(path[:-1])
            self._subboards[new_parent_board.path][path[-1]] = self._subboards[
                parent_board.path
            ].pop(replace_path[-1])
            self._subboards[replace_path] = {}

            if are_subboards_fetched is not None:
                self._are_subboards_fetched[replace_path] = are_subboards_fetched
            else:
                self._are_subboards_fetched[replace_path] = False

            return self._subboards[new_parent_board.path][path[-1]]
        else:
            # We use self.root's type because it may be a subclass of Board.
            self._subboards[parent_board.path][replace_path[-1]] = type(self.root)(
                path=replace_path, **kwargs
            )
            self._subboards[replace_path] = {}
            self._boards.append(self._subboards[parent_board.path][replace_path[-1]])

            if are_subboards_fetched is not None:
                self._are_subboards_fetched[replace_path] = are_subboards_fetched
            else:
                self._are_subboards_fetched[replace_path] = False

            return self._subboards[parent_board.path][replace_path[-1]]

    @final
    def _fetch_lower_boards(self, board: Board):
        if self._are_all_boards_fetched:
            return

        i = 0
        while i < len(self._boards):
            cur_board = self._boards[i]

            if cur_board.path[: len(board.path)] == board.path:
                self._fetch_subboards(cur_board)

            i += 1

        if not board.path:
            self._are_all_boards_fetched = True

    @final
    def _fetch_subboards(self, board: Board):
        try:
            self._do_fetch_subboards(board)
        except Exception as e:
            logging.warning(repr(e))
            logging.warning(traceback.format_exc())

    @abstractmethod
    def _do_fetch_subboards(self, board: Board):
        pass

    def _resolve_url(self, url: str):
        return url

    @abstractmethod
    def _get_node_from_url(self, url: str) -> Item | tuple[str, ...]:
        pass

    @final
    def find_board(self, path: tuple[str, ...]):
        # For now, `find_board` always needs all subboards to be fetched.
        if not self._are_all_boards_fetched:
            self._fetch_lower_boards(self.root)

        return self._find_board(path)

    @final
    def _find_board(self, path: tuple[str, ...]):
        cur_board: Board = self.root

        for path_part in path:
            if path_part not in self._subboards[cur_board.path]:
                self._fetch_lazy_subboard(cur_board, path_part)

            cur_board = self._subboards[cur_board.path][path_part]

        return cur_board

    @final
    def find_board_from_urls(self, urls: tuple[str, ...]):
        cur_board: Board = self.root

        for url in urls:
            subboard_urls = [
                subboard.url for _, subboard in self._subboards[cur_board.path].items()
            ]

            if url not in subboard_urls:
                self._fetch_lazy_subboards(cur_board)

            for _, subboard in self._subboards[cur_board.path].items():
                if subboard.url == url:
                    cur_board = subboard

        return cur_board

    @final
    def node_from_url(self, url: str):
        node = self._get_node_from_url(self._resolve_url(url))

        if isinstance(node, Board):
            return self.find_board(node.path)

        return node

    def _fetch_lazy_subboard(self, board: Board, subboard_id: str) -> Board | None:
        if not self._are_subboards_fetched[board.path]:
            for _ in self._fetch_lazy_subboards(board):
                pass

            self._are_subboards_fetched[board.path] = True

        return self._subboards[board.path][subboard_id]

    @abstractmethod
    def _fetch_lazy_subboards(self, board: Board) -> Generator[Board, None, None]:
        self._do_fetch_subboards(board)
        return iter(self._subboards[board.path])

    @final
    def board_count(self):
        return len(self._boards)

    @final
    def subboards(self, board: Board):
        if not self._are_subboards_fetched[board.path]:
            for _ in self._fetch_lazy_subboards(board):
                pass

            self._are_subboards_fetched[board.path] = True

        return self._subboards[board.path]

    @abstractmethod
    def _fetch_board_page_threads(
        self, board: Board, state: PageState
    ) -> Generator[Thread | File, None, PageState | None]:
        pass

    @final
    def _fetch_board_threads(
        self, board: Board, initial_state: PageState | None = None
    ):
        try:
            self.board_state = initial_state or PageState(url=board.url, page=1)
            while self.board_state:
                self.board_state = yield from self._fetch_board_page_threads(
                    board, self.board_state
                )
        except Exception as e:
            logging.warning(repr(e))
            logging.warning(traceback.format_exc())

    @abstractmethod
    def _fetch_thread_page_posts(
        self, thread: Thread, state: PageState
    ) -> Generator[Post | File, None, PageState | None]:
        pass

    @final
    def _fetch_thread_posts(
        self, thread: Thread, initial_state: PageState | None = None
    ):
        try:
            self.thread_state = initial_state or PageState(url=thread.url, page=1)
            while self.thread_state:
                self.thread_state = yield from self._fetch_thread_page_posts(
                    thread, self.thread_state
                )
        except Exception as e:
            logging.warning(repr(e))
            logging.warning(traceback.format_exc())

    @final
    def threads(self, board: Board, initial_state: PageState | None = None):
        for item in self._fetch_board_threads(board, initial_state):
            # if isinstance(item, Thread):
            match item:
                case Thread():
                    yield item
                case File():
                    pass

    @final
    def threads_with_files(self, board: Board, initial_state: PageState | None = None):
        yield from self._fetch_board_threads(board, initial_state)

    @final
    def posts(self, thread: Thread, initial_state: PageState | None = None):
        for item in self._fetch_thread_posts(thread, initial_state):
            match item:
                case Post():
                    yield item
                case File():
                    pass

    @final
    def posts_with_files(self, thread: Thread, initial_state: PageState | None = None):
        yield from self._fetch_thread_posts(thread, initial_state)

    @final
    def download_file(self, file: File):
        try:
            return self._session.try_get(file.url, should_cache=True)
        except Exception as e:
            logging.warning(repr(e))
            logging.warning(traceback.format_exc())


class HtmlExtractor(Extractor):
    _board_item_css: str
    _board_next_page_css: str
    _thread_item_css: str
    _thread_next_page_css: str

    @final
    def _fetch_board_page_threads(self, board: Board, state: PageState):
        response = self._session.get(state.url)
        soup = Soup(response.content)

        for tag in soup.soup.select(self._board_item_css):
            if thread := self._extract_board_page_thread(
                board, state, response, SoupTag(tag)
            ):
                yield thread

        yield from self._extract_file_objects((), (), soup, response)
        return self._extract_board_next_page_state(board, state, response, soup)

    @abstractmethod
    def _extract_board_page_thread(
        self, board: Board, state: PageState, response: Response, tag: SoupTag
    ) -> Thread | None:
        pass

    def _extract_board_next_page_state(
        self, board: Board, state: PageState, response: Response, soup: Soup
    ):
        if next_page_tag := soup.soup.select_one(self._board_next_page_css):
            href = cast(str, next_page_tag.get("href"))

            return PageState(url=urljoin(response.url, href), page=state.page + 1)

    @final
    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        response = self._session.get(state.url)
        soup = Soup(response.content)

        content_file_urls: list[str] = []

        for tag in soup.soup.select(self._thread_item_css):
            if post := self._extract_thread_page_post(
                thread, state, response, SoupTag(tag)
            ):
                yield post

                # TODO: Don't reparse text.
                new_content_file_urls = yield from self._extract_file_objects(
                    post.path, post.subpath, Soup(post.content), response
                )
                content_file_urls.extend(new_content_file_urls)

        for file in self._extract_file_objects((), (), soup, response):
            if file.url not in content_file_urls:
                yield file

        return self._extract_thread_next_page_state(thread, state, response, soup)

    @abstractmethod
    def _extract_thread_page_post(
        self, thread: Thread, state: PageState, response: Response, tag: SoupTag
    ) -> Post | None:
        pass

    def _extract_thread_next_page_state(
        self, thread: Thread, state: PageState, response: Response, soup: Soup
    ):
        if next_page_tag := soup.soup.select_one(self._thread_next_page_css):
            href = cast(str, next_page_tag.get("href"))

            if not href:
                return

            return PageState(url=urljoin(response.url, href), page=state.page + 1)

    @final
    def _extract_file_objects(
        self,
        path: tuple[str, ...],
        subpath: tuple[str, ...],
        soup_or_tag: Soup | SoupTag,
        response: Response,
    ):
        match soup_or_tag:
            case Soup():
                obj = soup_or_tag.soup
            case SoupTag():
                obj = soup_or_tag.tag

        embeds = obj.select(
            'link[rel="stylesheet"], embed, audio, img, object, svg, video'
        )

        urls: list[str] = []

        for embed in embeds:
            embed = SoupTag(embed)
            url = None

            if embed.tag.name == "link":
                url = urljoin(response.url, embed.get("href"))
                yield File(
                    path=path,
                    url=url,
                    origin=response.url,
                    data={},
                    subpath=subpath + (url,),
                )
            elif embed.tag.name == "embed":
                url = urljoin(response.url, embed.get("src"))
                yield File(
                    path=path,
                    url=url,
                    origin=response.url,
                    data={},
                    subpath=subpath + (url,),
                )
            elif embed.tag.name == "audio":
                for source in embed.tag.find_all("source"):
                    url = urljoin(response.url, source.get("src"))
                    yield File(
                        path=path,
                        url=url,
                        origin=response.url,
                        data={},
                        subpath=subpath + (url,),
                    )
            elif embed.tag.name == "img":
                url = urljoin(response.url, embed.get("src"))
                yield File(
                    path=path,
                    url=url,
                    origin=response.url,
                    data={},
                    subpath=subpath + (url,),
                )
            elif embed.tag.name == "object":
                url = urljoin(response.url, embed.get("data"))
                yield File(
                    path=path,
                    url=url,
                    origin=response.url,
                    data={},
                    subpath=subpath + (url,),
                )
            elif embed.tag.name == "svg":
                yield File(
                    path=path,
                    url=response.url,
                    origin=response.url,
                    data={},
                    subpath=subpath,
                    content_type="image/svg+xml",
                    content=embed.encode_contents(),
                )

            if url:
                urls.append(url)

        return urls



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from urllib.parse import urljoin
import dateutil.parser
import re

from .common import normalize_url, regex_match, regex_search
from .common import HtmlExtractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup, SoupTag

if TYPE_CHECKING:
    from requests import Response


class SimplemachinesExtractor(HtmlExtractor):
    tests = [
        {
            "url": "https://simplemachines.org/community",
            "test_base_url": "https://www.simplemachines.org/community/",
            "test_boards": {
                ("2",): {
                    "title": "Simple Machines",
                },
                ("2", "1"): {
                    "title": "News and Updates",
                },
                ("2", "244"): {
                    "title": "Organizational News and Updates",
                },
                ("3",): {
                    "title": "SMF Support",
                },
                ("3", "254"): {
                    "title": "SMF 2.1.x Support",
                },
                ("3", "254", "255"): {
                    "title": "PostgreSQL Support",
                },
                ("3", "147"): {
                    "title": "SMF 2.0.x Support",
                },
                ("3", "147", "222"): {
                    "title": "PostgreSQL and SQLite Support",
                },
                ("3", "9"): {
                    "title": "SMF 1.1.x Support",
                },
                ("3", "12"): {
                    "title": "Language Specific Support",
                },
                ("3", "20"): {
                    "title": "Converting to SMF",
                },
                ("3", "20", "132"): {
                    "title": "IPB",
                },
                ("3", "20", "11"): {
                    "title": "MyBB",
                },
                ("3", "20", "133"): {
                    "title": "phpBB",
                },
                ("3", "20", "134"): {
                    "title": "vBulletin",
                },
                ("3", "20", "135"): {
                    "title": "YaBB/YaBB SE",
                },
                ("18",): {
                    "title": "Customizing SMF",
                },
                ("18", "60"): {
                    "title": "SMF Coding Discussion",
                },
                ("18", "33"): {
                    "title": "Bridges and Integrations",
                },
                ("18", "59"): {
                    "title": "Modifications and Packages",
                },
                ("18", "59", "79"): {
                    "title": "Mod Requests",
                },
                ("18", "34"): {
                    "title": "Graphics and Templates",
                },
                ("18", "34", "106"): {
                    "title": "Theme Site Themes",
                },
                ("18", "34", "178"): {
                    "title": "Theme Previews",
                },
                ("18", "72"): {
                    "title": "Tips and Tricks",
                },
                ("18", "72", "115"): {
                    "title": "Now Available",
                },
                ("18", "116"): {
                    "title": "Building Your Community and other Forum Advice",
                },
                ("15",): {
                    "title": "SMF Development",
                },
                ("15", "3"): {
                    "title": "Feature Requests",
                },
                ("15", "3", "38"): {
                    "title": "Applied or Declined Requests",
                },
                ("15", "3", "228"): {
                    "title": "Next SMF Discussion",
                },
                ("15", "137"): {
                    "title": "Bug Reports",
                },
                ("15", "137", "37"): {
                    "title": "Fixed or Bogus Bugs",
                },
                ("15", "137", "197"): {
                    "title": "Bugtracker (Github)",
                },
                ("4",): {
                    "title": "General Community",
                },
                ("4", "19"): {
                    "title": "Site Comments, Issues and Concerns",
                },
                ("4", "8"): {
                    "title": "Scripting Help",
                },
                ("4", "7"): {
                    "title": "Test Board",
                },
                ("16",): {
                    "title": "Simple Machines Blogs",
                },
                ("16", "128"): {
                    "title": "SMF Team Blog",
                },
                ("16", "129"): {
                    "title": "Developers' Blog",
                },
                ("5",): {
                    "title": "Archived Boards and Threads...",
                },
                ("5", "136"): {
                    "title": "Archived Boards",
                },
                ("5", "136", "2"): {
                    "title": "SMF Feedback and Discussion",
                },
                ("5", "136", "41"): {
                    "title": "Parham's PHP Tutorials",
                },
                ("5", "136", "96"): {
                    "title": "Classic Themes",
                },
                ("5", "136", "144"): {
                    "title": "MotM Travel Blog",
                },
                ("5", "136", "77"): {
                    "title": "Mambo Bridge Support",
                },
                ("5", "136", "138"): {
                    "title": "Joomla Bridge Support",
                },
                ("5", "136", "10"): {
                    "title": "Install and Upgrade Help",
                },
            },
        },
        {
            "url": "https://www.simplemachines.org/community/index.php?board=255.0",
            "test_base_url": "https://www.simplemachines.org/community/",
            "test_boards": {
                ("3", "254", "255"): {
                    "title": "PostgreSQL Support",
                },
                ("3", "20", "134"): {
                    "title": "vBulletin",
                },
                ("4", "19"): {
                    "title": "Site Comments, Issues and Concerns",
                },
            },
        },
        {
            "url": "https://www.simplemachines.org/community/index.php?topic=573.0",
            "test_base_url": "https://www.simplemachines.org/community/",
            "test_contents_hash": "daaf0300248ec6a1891aa1498c6637a89769dd7e",
            "test_item_count": 6,
        },
        {
            "url": "https://www.simplemachines.org/community/index.php?topic=197.0",
            "test_base_url": "https://www.simplemachines.org/community/",
            "test_contents_hash": "cd53463ffb3965590f925807902b6962893a1d5a",
            "test_item_count": 50,
        },
        {
            "url": "https://www.eevblog.com/forum/cooking/",
            "test_base_url": "https://www.eevblog.com/forum/",
            "test_min_item_count": 150,
        },
        {
            "url": "https://www.eevblog.com/forum/eda/best-sub-$2k-pcb-design-software/",
            "test_base_url": "https://www.eevblog.com/forum/",
            # EEVBlog forum uses a PHPSESSID query for tracking, which makes the hash unpredictable.
            # So, we don't test the hash.
            # "test_contents_hash": "f9288cedb415d05e70887e658234a3c0976990c6",
            "test_item_count": 69,
        },
        {
            "url": "http://www.bay12forums.com/smf/index.php?board=10.0",
            "test_base_url": "http://www.bay12forums.com/smf/",
            "test_min_item_count": 150,
        },
        {
            "url": "http://www.bay12forums.com/smf/index.php?topic=406.0",
            "test_base_url": "http://www.bay12forums.com/smf/",
            "test_item_count": 31,
            "test_contents_hash": "4e2c2d065d83c23592379ef97c6d48dbea4f9f4f",
        },
    ]

    _board_item_css = 'span[id^="msg_"]'
    _thread_item_css = "div.post_wrapper, :has(> .postarea)"
    _board_next_page_css = 'a.nav_page:has(span.next_page), a.navPages:-soup-contains("Next"), strong + a.navPages'
    _thread_next_page_css = 'a.nav_page:has(span.next_page), a.navPages:-soup-contains("Next"), strong + a.navPages'

    _category_id_regex = re.compile(r"^c(\d+)$")
    _board_id_regex = re.compile(r"^b(\d+)$")
    _div_id_regex = re.compile(r"^msg_(\d+)$")
    _span_id_regex = re.compile(r"^msg_(\d+)$")
    _subject_id_regex = re.compile(r"^subject_(\d+)$")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.try_get(url, should_cache=True, should_retry=False)
        soup = Soup(response.content)

        link = soup.find("link", attrs={"rel": "contents"})
        base_url = normalize_url(link.get("href"))

        simplemachines_anchor = soup.find(
            "a",
            attrs={
                "href": re.compile(r"https?://www.simplemachines.org"),
                "title": "Simple Machines",
            },
        )

        if simplemachines_anchor:
            return SimplemachinesExtractor(session, base_url, options)

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url, should_cache=True)
        soup = Soup(response.content)

        category_anchors = soup.find_all("a", id=self._category_id_regex)
        for category_anchor in category_anchors:
            category_id = regex_match(
                self._category_id_regex, category_anchor.get("id")
            ).group(1)
            category_title = str(category_anchor.next_sibling).strip()

            self._set_board(
                path=(category_id,),
                url=urljoin(response.url, f"index.php#c{category_id}"),
                origin=response.url,
                data={},
                title=category_title,
                are_subboards_fetched=True,
            )

            for parent in category_anchor.parents:
                board_anchors = parent.find_all("a", id=self._board_id_regex)

                if not board_anchors:
                    board_anchors = parent.find_all(
                        "a", attrs={"name": self._board_id_regex}
                    )

                if board_anchors:
                    for board_anchor in board_anchors:
                        if board_anchor.get("id"):
                            board_id = regex_match(
                                self._board_id_regex, board_anchor.get("id")
                            ).group(1)
                        else:
                            board_id = regex_match(
                                self._board_id_regex, board_anchor.get("name")
                            ).group(1)

                        self._set_board(
                            path=(category_id, board_id),
                            url=board_anchor.get("href"),
                            origin=response.url,
                            data={},
                            title=str(board_anchor.string).strip(),
                            are_subboards_fetched=True,
                        )
                    break

    def _do_fetch_subboards(self, board: Board):
        if not board.url:
            return

        # Don't fetch top boards.
        if len(board.path) <= 1:
            return

        response = self._session.get(board.url, should_cache=True)
        soup = Soup(response.content)

        subboard_anchors = soup.find_all("a", attrs={"id": self._board_id_regex})

        for subboard_anchor in subboard_anchors:
            subboard_id = regex_match(
                self._board_id_regex, subboard_anchor.get("id")
            ).group(1)
            self._set_board(
                path=board.path + (subboard_id,),
                url=subboard_anchor.get("href"),
                origin=response.url,
                data={},
                title=str(subboard_anchor.string).strip(),
                are_subboards_fetched=True,
            )

    def _resolve_url(self, url: str):
        return normalize_url(
            self._session.get(url, should_cache=True).url,
            append_slash=True,
            keep_queries=["board", "topic"],
        )

    def _get_node_from_url(self, url: str):
        response = self._session.get(url, should_cache=True)
        soup = Soup(response.content)

        breadcrumbs = soup.try_find(class_="navigate_section")
        if not breadcrumbs:
            breadcrumbs = soup.find(class_="linktree")

        breadcrumb_lis = breadcrumbs.find_all("li")
        breadcrumb_as = [li.find("a") for li in breadcrumb_lis]

        # Thread.
        if soup.try_find("div", id="forumposts"):
            breadcrumb_urls = [a.get("href") for a in breadcrumb_as]
            board = self.find_board_from_urls(tuple(breadcrumb_urls[1:-1]))

            topic_input = soup.find("input", attrs={"name": "topic"})
            thread_id = topic_input.get("value")
            title_title = soup.find("title")

            return Thread(
                path=board.path + (thread_id,),
                url=url,
                origin=response.url,
                data={},
                title=str(title_title.string),
            )

        # Board.
        else:
            self._fetch_lower_boards(self.root)

            board_href = self._resolve_url(breadcrumb_as[-1].get("href"))

            for cur_board in self._boards:
                if cur_board.url == board_href:
                    return cur_board

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, subboard_id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _extract_board_page_thread(
        self, board: Board, state: PageState, response: Response, tag: SoupTag
    ):
        thread_id = regex_match(self._span_id_regex, tag.get("id")).group(1)
        msg_anchor = tag.tags[0]

        return Thread(
            path=board.path + (thread_id,),
            url=msg_anchor.get("href"),
            origin=response.url,
            data={},
            title=str(msg_anchor.string),
        )

    def _extract_thread_page_post(
        self, thread: Thread, state: PageState, response: Response, tag: SoupTag
    ):
        msg_div = tag.find("div", id=self._div_id_regex)
        subject_tag = tag.find({"h5", "div"}, id=self._subject_id_regex)

        time_tag = subject_tag.find_next({"a", "div"}, class_="smalltext")

        # This is ugly, but it's the best I can do for now.
        date = regex_search(
            re.compile(
                r"(January|February|March|April|May|June|July|August|September|October|November|December|Yesterday|Today) [a-zA-Z0-9,: ]+"
            ),
            time_tag.tag.get_text(),  # Get rid of HTML tags.
        ).group(0)

        poster_div = tag.find("div", class_="poster")
        poster_h4 = poster_div.find("h4")

        if poster_anchor := poster_h4.try_find("a"):
            author = poster_anchor.string
        else:
            author = poster_h4.string.strip()

        return Post(
            path=thread.path,
            subpath=(regex_match(self._div_id_regex, msg_div.get("id")).group(1),),
            url=subject_tag.find("a").get("href"),
            origin=response.url,
            data={},
            author=author,
            creation_time=dateutil.parser.parse(date, fuzzy=True).isoformat(),
            content="".join(str(v) for v in msg_div.contents).strip(),
        )



# type: ignore
from __future__ import annotations
from typing import *  # type: ignore

from pydantic import BaseModel
from functools import lru_cache, wraps
from tenacity import (
    retry,
    wait_random_exponential,
    stop_after_attempt,
    before_sleep_log,
)
import time
import logging

from .exceptions import CrawlError
from .version import __version__

if TYPE_CHECKING:
    from requests import Response


class SessionOptions(BaseModel):
    timeout: float
    retries: int
    retry_sleep: float
    retry_sleep_multiplier: float
    warc_output: str
    user_agent: str
    get_urls: bool


class Session:
    def __init__(self, options: SessionOptions):
        self._warc_file = None

        if options.warc_output:
            from warcio.warcwriter import WARCWriter
            from warcio.capture_http import capture_http

            self._capture_http = capture_http
            self._warc_file = open(options.warc_output, "wb")
            self._warc_writer = WARCWriter(self._warc_file)

        # For the `warcio` recording to work, `requests` must be imported only after `capture_http`.
        import requests

        self._session = requests.Session()
        self._options = options
        self._cache: dict[
            tuple[str, frozenset[tuple[str, Any]], frozenset[tuple[str, Any]]],
            requests.Response,
        ] = {}
        self._past_requests: set[
            tuple[str, frozenset[tuple[str, Any]], frozenset[tuple[str, Any]]]
        ] = set()

        self.delay = 1
        self.attempts = 0

    def __del__(self):
        if self._warc_file:
            self._warc_file.close()

    def get(
        self,
        url: str,
        *,
        params: dict[str, Any] = {},
        headers: dict[str, Any] = {},
        should_cache: bool = False,
        should_retry: bool = True,
        **kwargs: Any,
    ):
        response = self.try_get(
            url,
            params=params,
            headers=headers,
            should_cache=should_cache,
            should_retry=should_retry,
            **kwargs,
        )
        response.raise_for_status()

        return response

    def try_get(
        self,
        url: str,
        *,
        params: dict[str, Any] = {},
        headers: dict[str, Any] = {},
        should_cache: bool = False,
        should_retry: bool = True,
        **kwargs: Any,
    ) -> Response:
        logging.debug(f"Attempting GET {url} {params} {headers}")

        frozen_params = frozenset(params.items())
        frozen_headers = frozenset(headers.items())

        if (url, frozen_params, frozen_headers) in self._cache:
            cached_response = self._cache[(url, frozen_params, frozen_headers)]

            if not should_cache:
                del self._cache[(url, frozen_params, frozen_headers)]

            return cached_response
        elif (url, frozen_params, frozen_headers) in self._past_requests:
            raise CrawlError(url, frozen_params, frozen_headers)

        if should_retry:

            @retry(
                reraise=True,
                wait=wait_random_exponential(
                    multiplier=self._options.retry_sleep,
                    exp_base=self._options.retry_sleep_multiplier,
                ),
                stop=stop_after_attempt(self._options.retries),
                before_sleep=before_sleep_log(logging.getLogger(), logging.WARNING),
            )
            def retrying_get(
                url: str,
                *,
                params: dict[str, Any] = {},
                headers: dict[str, Any] = {},
                **kwargs: Any,
            ):
                return self._do_get(url, params=params, headers=headers, **kwargs)

            response = retrying_get(url, params=params, headers=headers, **kwargs)
        else:
            response = self._do_get(url, params=params, headers=headers, **kwargs)

        if should_cache:
            self._cache[(url, frozen_params, frozen_headers)] = response
        else:
            self._past_requests.add((url, frozen_params, frozen_headers))

        return response

    def _after_retry(self):
        logging.warning(f"Waiting {self.delay} seconds.")

    def _do_get(
        self,
        url: str,
        *,
        params: dict[str, Any] = {},
        headers: dict[str, Any] = {},
        **kwargs: Any,
    ):
        if self._options.get_urls:
            print(url)
        else:
            logging.info(f"GET {url} {params} {headers}")

        if not headers:
            headers = {"User-Agent": self._options.user_agent}

        if self._warc_file:
            with self._capture_http(self._warc_writer):
                return self._session.get(
                    url,
                    params=params,
                    headers=headers,
                    timeout=self._options.timeout,
                    **kwargs,
                )
        else:
            return self._session.get(
                url,
                params=params,
                headers=headers,
                timeout=self._options.timeout,
                **kwargs,
            )

    def validate_url(self, url: str):
        try:
            self._session.get_adapter(url)
        except:  # `InvalidSchema`, not referencing it directly to avoid breaking WARC recording.
            return False

        return True



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from abc import ABC, abstractmethod
from pydantic import BaseModel
from mailbox import Mailbox, Message
from urllib.parse import urlparse
from base64 import b64encode, b64decode
from urllib.parse import quote_plus
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.encoders import encode_base64

import email.utils
import os
import re

try:
    from html2text import html2text
except ImportError:
    pass

from datetime import datetime, timezone
import sys

from ..extractors.common import Extractor, Item, Thread, Board, Post, File, PageState
from ..version import __version__


class WriterOptions(BaseModel):
    output_path: str
    files_output_path: str
    write_board_objects: bool
    write_thread_objects: bool
    write_post_objects: bool
    write_file_objects: bool
    write_outside_file_objects: bool
    textify: bool
    content_as_title: bool
    author_as_addr_spec: bool


class WriterState(BaseModel):
    board_path: tuple[str, ...] | None = None
    board_page: PageState | None = None
    thread_page: PageState | None = None


class Entry(BaseModel):
    generator: str
    version: str
    extractor: str
    download_time: str
    type: str
    item: Item

    class Config:
        json_encoders: dict[Any, Callable[[Any], str]] = {
            bytes: lambda content: b64encode(content).decode("ascii"),
        }


class Writer(ABC):
    tests: list[dict[str, Any]]

    def __init__(self, extractor: Extractor, options: WriterOptions):
        self._extractor = extractor
        self._options = options
        self._initial_state = WriterState()

    def write(self, url: str):
        self.read_metadata()

        base_node = self._extractor.node_from_url(url)

        if isinstance(base_node, Board):
            self.write_board(base_node)
        elif isinstance(base_node, Thread):
            self.write_thread(base_node)

    @abstractmethod
    def read_metadata(self):
        pass

    @abstractmethod
    def _write_board_object(self, board: Board):
        pass

    def _write_board_threads(self, board: Board):
        for item in self._extractor.threads_with_files(
            board, self._initial_state.board_page
        ):
            match item:
                case Thread():
                    self.write_thread(item)
                case File():
                    self.write_file(item)

    @final
    def write_board(self, board: Board):
        if self._options.write_board_objects:
            self._write_board_object(board)

        self._write_board_threads(board)

        for _, subboard in self._extractor.subboards(board).items():
            self.write_board(subboard)

    @abstractmethod
    def _write_thread_object(self, thread: Thread):
        pass

    def _write_thread_posts(self, thread: Thread):
        for item in self._extractor.posts_with_files(
            thread, self._initial_state.thread_page
        ):
            match item:
                case Post():
                    self.write_post(thread, item)
                case File():
                    self.write_file(item)

    @final
    def write_thread(self, thread: Thread):
        if self._options.write_thread_objects:
            self._write_thread_object(thread)

        self._write_thread_posts(thread)

    @abstractmethod
    def _write_post_object(self, thread: Thread, post: Post):
        pass

    @final
    def write_post(self, thread: Thread, post: Post):
        if self._options.write_post_objects:
            self._write_post_object(thread, post)

    @final
    def write_file(self, file: File):
        if not self._options.write_file_objects:
            return

        if not file.path and not self._options.write_outside_file_objects:
            return

        if self._options.files_output_path:
            os.makedirs(self._options.files_output_path, exist_ok=True)

            file_path = os.path.join(
                self._options.files_output_path, quote_plus(file.url)
            )

            if file.content:
                file.os_path = file_path

                with open(file_path, "wb") as f:
                    f.write(file.content)

                file.content = None
            elif match := re.match("data:(.+/.+);base64,(.*)", file.url):
                file.content_type = match.group(1)
                file.os_path = file_path

                with open(file_path, "wb") as f:
                    f.write(b64decode(match.group(2)))
            elif response := self._extractor.download_file(file):
                print(response.headers)
                file.content_type = response.headers.get(
                    "Content-Type", "application/octet-stream"
                )
                file.os_path = file_path

                with open(file_path, "wb") as f:
                    f.write(response.content)
        else:
            if file.content:
                file.content = b64encode(file.content)
            elif match := re.match("data:(.+/.+);base64,(.*)", file.url):
                file.content_type = match.group(1)
            else:
                if response := self._extractor.download_file(file):
                    file.content_type = response.headers.get(
                        "Content-Type", "application/octet-stream"
                    )
                    file.content = b64encode(response.content)

        self._write_file_object(file)

    @abstractmethod
    def _write_file_object(self, file: File):
        pass


class SimulatedWriter(Writer):
    def read_metadata(self):
        pass

    def _write_board_object(self, board: Board):
        pass

    def _write_thread_object(self, thread: Thread):
        pass

    def _write_post_object(self, thread: Thread, post: Post):
        pass

    def _write_file_object(self, file: File):
        pass


class FileWriter(Writer):
    def __init__(self, extractor: Extractor, options: WriterOptions):
        super().__init__(extractor, options)

        if options.output_path != "-":
            self._file = open(options.output_path, "w")
        else:
            self._file = None

    def __del__(self):
        if self._file:
            self._file.close()

    def read_metadata(self):
        pass  # TODO.

    def _write_board_object(self, board: Board):
        entry = self._make_entry(board)

        if self._file:
            self._file.write(f"{self._serialize_entry(entry)}\n")
        else:
            sys.stdout.write(f"{self._serialize_entry(entry)}\n")

    def _write_thread_object(self, thread: Thread):
        entry = self._make_entry(thread)

        if self._file:
            self._file.write(f"{self._serialize_entry(entry)}\n")
        else:
            sys.stdout.write(f"{self._serialize_entry(entry)}\n")

    def _write_post_object(self, thread: Thread, post: Post):
        entry = self._make_entry(post)

        if self._file:
            self._file.write(f"{self._serialize_entry(entry)}\n")
        else:
            sys.stdout.write(f"{self._serialize_entry(entry)}\n")

    def _write_file_object(self, file: File):
        entry = self._make_entry(file)

        if self._file:
            self._file.write(f"{self._serialize_entry(entry)}\n")
        else:
            sys.stdout.write(f"{self._serialize_entry(entry)}\n")

    def _make_entry(self, item: Item):
        match item:
            case Board():
                typ = "board"
            case Thread():
                typ = "thread"
            case Post():
                typ = "post"
            case File():
                typ = "file"
            case Item():
                raise ValueError

        return Entry(
            generator="forum-dl",
            version=__version__,
            extractor=self._extractor.__class__.__module__.split(".")[-1],
            download_time=datetime.now(timezone.utc).isoformat(),
            type=typ,
            item=item,
        )

    @abstractmethod
    def _serialize_entry(self, entry: Entry) -> str:
        pass


class MailWriter(Writer):
    def __init__(
        self,
        extractor: Extractor,
        mailbox: Mailbox[Any],
        options: WriterOptions,
    ):
        super().__init__(extractor, options)
        self._mailbox = mailbox
        self._message_key = None

        for key, msg in self._mailbox.iteritems():
            if msg.get("X-Forumdl-Version"):
                self._metadata_key = key
        else:
            msg = self._new_message()
            self._metadata_key = self._mailbox.add(msg)

    def __del__(self):
        self._mailbox.flush()
        self._mailbox.close()

    def write(self, url: str):
        self._mailbox.lock()
        super().write(url)
        self._mailbox.unlock()

    def read_metadata(self):
        pass  # TODO.

    def _write_board_object(self, board: Board):
        pass  # TODO.

    def _write_thread_object(self, thread: Thread):
        pass  # TODO.

    def _write_post_object(self, thread: Thread, post: Post):
        self._post = post
        self._message_key = self._mailbox.add(self._build_message(thread, post))

    def _write_file_object(self, file: File):
        if file.subpath[:-1] == self._post.subpath:
            part = Message()
            part["Content-Type"] = file.content_type
            part["MIME-Version"] = "1.0"

            if response := self._extractor.download_file(file):
                part.set_payload(response.content)

            encode_base64(part)

            part.add_header(
                f"Content-Disposition",
                f"attachment; filename={quote_plus(file.url)}",
            )
            self._attach_part(part)

    def _attach_part(self, part: Message):
        if self._message_key:
            msg = self._mailbox[self._message_key]
            msg.attach(part)
            self._mailbox[self._message_key] = msg

    @abstractmethod
    def _new_message(self) -> Message:
        pass

    def _build_message(self, thread: Thread, post: Post):
        msg = self._new_message()
        msg.set_payload([])

        path = post.path + post.subpath

        msg["Message-ID"] = "<" + ".".join(path) + ">"
        msg["Content-Location"] = post.url
        msg["Date"] = email.utils.formatdate(
            datetime.fromisoformat(post.creation_time).timestamp()
        )

        if self._options.author_as_addr_spec:
            domain = urlparse(self._extractor.base_url).netloc

            msg["From"] = f"{post.author} <{post.author}@{domain}>"
        else:
            msg["From"] = post.author

        if len(path) >= 2:
            msg["In-Reply-To"] = f"<{'.'.join(path[:-1])}>"

            refs = f"{path[0]}"
            for ref in post.path[1:-1]:
                refs += f" <{ref}>"

        if len(post.subpath) >= 1 and self._options.content_as_title:
            msg["Subject"] = html2text(post.content[:98]).partition("\n")[0]
        else:
            msg["Subject"] = thread.title

        # msg["Date"] = formatdate(post.date)

        alternativeMsg = MIMEMultipart("alternative")
        msg.attach(alternativeMsg)

        if self._options.textify:
            alternativeMsg.attach(MIMEText(html2text(post.content)))
        else:
            alternativeMsg.attach(MIMEText(post.content, "html"))

        return msg


class FolderedMailWriter(MailWriter):
    def __init__(
        self,
        extractor: Extractor,
        mailbox: Mailbox[Any],
        options: WriterOptions,
    ):
        super().__init__(extractor, mailbox, options)
        self.folders: dict[str, Mailbox[Any]] = {}
        self._folder_name = None

    def _get_folder_name(self, board: Board):
        return ".".join(board.path)

    def _write_board_object(self, board: Board):
        folder_name = self._get_folder_name(board)

        if folder_name not in self.folders:
            self.folders[folder_name] = getattr(self._mailbox, "add_folder")(
                folder_name
            )

        super()._write_board_object(board)

    def _write_post_object(self, thread: Thread, post: Post):
        board = self._extractor.find_board(thread.path[:-1])
        folder_name = self._get_folder_name(board)

        if folder_name not in self.folders:
            self.folders[folder_name] = getattr(self._mailbox, "add_folder")(
                folder_name
            )

        self._post = post
        self._folder_name = folder_name
        self._message_key = self.folders[folder_name].add(
            self._build_message(thread, post)
        )

    def _attach_part(self, part: Message):
        if self._folder_name and self._message_key:
            msg = self.folders[self._folder_name][self._message_key]
            msg.attach(part)
            self.folders[self._folder_name][self._message_key] = msg



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

import argparse
import logging

from .version import __version__


def build_parser():
    parser = argparse.ArgumentParser(
        add_help=False,
    )

    general = parser.add_argument_group("General Options")
    general.add_argument(
        "--help",
        action="help",
        help="Show this help message and exit",
    )
    general.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print program version and exit",
    )
    general.add_argument(
        "--list-extractors",
        dest="list_extractors",
        action="store_true",
        help="List all supported extractors and exit",
    )
    general.add_argument(
        "--list-output-formats",
        dest="list_output_formats",
        action="store_true",
        help="List all supported output formats and exit",
    )

    session = parser.add_argument_group("Session Options")
    session.add_argument(
        "--timeout",
        metavar="SECONDS",
        dest="timeout",
        default="5",
        help="HTTP connection timeout",
    )
    session.add_argument(
        "-R",
        "--retries",
        metavar="N",
        dest="retries",
        default="4",
        help="Maximum number of retries for failed HTTP requests or -1 to retry infinitely (default: 4)",
    )
    session.add_argument(
        "--retry-sleep",
        metavar="SECONDS",
        dest="retry_sleep",
        default="1",
        help="Time to sleep between retries, in seconds (default: 1)",
    )
    session.add_argument(
        "--retry-sleep-multiplier",
        metavar="K",
        dest="retry_sleep_multiplier",
        default="2",
        help="A constant by which sleep time is multiplied on each retry (default: 2)",
    )
    session.add_argument(
        "--user-agent",
        metavar="UA",
        dest="user_agent",
        default=f"Forum-dl {__version__}",
        help="User-Agent request header",
    )

    output = parser.add_argument_group("Output Options")
    output.add_argument(
        "-q",
        "--quiet",
        dest="loglevel",
        default=logging.INFO,
        action="store_const",
        const=logging.ERROR,
        help="Activate quiet mode",
    )
    output.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        action="store_const",
        const=logging.DEBUG,
        help="Print various debugging information",
    )
    output.add_argument(
        "-g",
        "--get-urls",
        dest="get_urls",
        action="store_true",
        help="Print URLs instead of downloading",
    )
    output.add_argument(
        "-o",
        "--output",
        metavar="OUTFILE",
        dest="output",
        default="-",
        help="Output all results concatenated to OUTFILE, or stdout if OUTFILE is - (default: -)",
    )
    output.add_argument(
        "-f",
        "--output-format",
        metavar="FORMAT",
        dest="output_format",
        default="jsonl",
        help="Output format. Use --list-output-formats for a list of possible arguments",
    )
    output.add_argument(
        "--warc-output",
        metavar="FILE",
        dest="warc_output",
        default="",
        help="Record HTTP requests, store them in FILE in WARC format",
    )
    output.add_argument(
        "--files-output",
        metavar="DIR",
        dest="files_output",
        default="",
        help="Store files in DIR instead of OUTFILE",
    )
    output.add_argument(
        "--boards",
        dest="boards",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="Write board objects (default: True, --no-boards to negate)",
    )
    output.add_argument(
        "--threads",
        dest="threads",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="Write thread objects (default: True, --no-threads to negate)",
    )
    output.add_argument(
        "--posts",
        dest="posts",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="Write post objects (default: True, --no-posts to negate)",
    )
    output.add_argument(
        "--files",
        dest="files",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="Write embedded files (--no-files to negate)",
    )
    output.add_argument(
        "--outside-files",
        dest="outside_files",
        default=False,
        action=argparse.BooleanOptionalAction,
        help="Write embedded files outside post content. Auto-enabled by --warc-output and -f warc (default: False, --no-outside-files to negate)",
    )
    output.add_argument(
        "--textify",
        dest="textify",
        action="store_true",
        help="Lossily convert HTML content to plaintext",
    )
    output.add_argument(
        "--content-as-title",
        dest="content_as_title",
        action="store_true",
        help="Write 98 initial characters of content in title field of each post",
    )
    output.add_argument(
        "--author-as-addr-spec",
        dest="author_as_addr_spec",
        action="store_true",
        help="Append author and domain as an addr-spec in the From header",
    )

    parser.add_argument(
        "urls",
        metavar="URL",
        nargs="*",
        help=argparse.SUPPRESS,
    )

    return parser



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from urllib.parse import urljoin
import re

from .common import normalize_url, regex_match
from .common import HtmlExtractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup, SoupTag

if TYPE_CHECKING:
    from requests import Response


class XenforoExtractor(HtmlExtractor):
    tests = [
        {
            "url": "https://xenforo.com/community",
            "test_base_url": "https://xenforo.com/community/",
            "test_boards": {
                ("1",): {
                    "title": "Official forums",
                },
                ("1", "2"): {
                    "title": "Announcements",
                },
                ("1", "3"): {
                    "title": "Have you seen...?",
                },
                ("1", "19"): {
                    "title": "Frequently asked questions",
                },
                ("17",): {
                    "title": "Public forums",
                },
                ("17", "5"): {
                    "title": "XenForo pre-sales questions",
                },
                ("17", "18"): {
                    "title": "XenForo suggestions",
                },
                ("46",): {
                    "title": "XenForo bug reports",
                },
                ("46", "91"): {
                    "title": "Bug reports",
                },
                ("46", "105"): {
                    "title": "Importer bug reports",
                },
                ("46", "92"): {
                    "title": "Resolved bug reports",
                },
                ("46", "43"): {
                    "title": "Future-fix bug reports",
                },
                ("22",): {
                    "title": "Customer forums",
                },
                ("22", "23"): {
                    "title": "Installation, upgrade, and import support",
                },
                ("22", "24"): {
                    "title": "Troubleshooting and problems",
                },
                ("22", "25"): {
                    "title": "XenForo questions and support",
                },
                ("22", "47"): {
                    "title": "Styling and customization questions",
                },
                ("22", "48"): {
                    "title": "Server configuration and hosting",
                },
                ("22", "84"): {
                    "title": "Forum management",
                },
                ("22", "79"): {
                    "title": "Tips and guides [1.x]",
                },
                ("22", "94"): {
                    "title": "Tips and guides [2.x]",
                },
                ("22", "45"): {
                    "title": "Other XenForo discussions and feedback",
                },
                ("22", "49"): {
                    "title": "XenForo help and manual",
                },
                ("54",): {
                    "title": "Official XenForo add-ons",
                },
                ("54", "86"): {
                    "title": "Media Gallery suggestions",
                },
                ("54", "87"): {
                    "title": "Media Gallery support",
                },
                ("54", "88"): {
                    "title": "Media Gallery bug reports",
                },
                ("54", "65"): {
                    "title": "Resource Manager suggestions",
                },
                ("54", "82"): {
                    "title": "Resource Manager support",
                },
                ("54", "66"): {
                    "title": "Resource Manager bug reports",
                },
                ("54", "55"): {
                    "title": "Enhanced Search suggestions",
                },
                ("54", "56"): {
                    "title": "Enhanced Search support",
                },
                ("54", "57"): {
                    "title": "Enhanced Search bug reports",
                },
                ("60",): {
                    "title": "XenForo resources and add-ons",
                },
                ("60", "70"): {
                    "title": "Resources and add-ons",
                },
                ("60", "71"): {
                    "title": "Resource and add-on releases [1.x]",
                },
                ("60", "93"): {
                    "title": "Resource and add-on releases [2.x]",
                },
                ("60", "68"): {
                    "title": "Resource and add-on requests",
                },
                ("60", "63"): {
                    "title": "Resource and add-on discussions",
                },
                ("60", "69"): {
                    "title": "Custom service/development requests",
                },
                ("60", "42"): {
                    "title": "Third-party services & offers",
                },
                ("60", "62"): {
                    "title": "Resource and add-on archive",
                },
                ("51",): {
                    "title": "Development help",
                },
                ("51", "106"): {
                    "title": "Resource standards",
                },
                ("51", "80"): {
                    "title": "Development tutorials and resources [1.x]",
                },
                ("51", "101"): {
                    "title": "Development tutorials [2.x]",
                },
                ("51", "34"): {
                    "title": "XenForo development discussions",
                },
                ("51", "52"): {
                    "title": "General PHP and MySQL discussions",
                },
                ("4",): {
                    "title": "General discussions",
                },
                ("4", "53"): {
                    "title": "Forum showcase and critiques",
                },
                ("4", "7"): {
                    "title": "Off topic",
                },
                ("14",): {
                    "title": "Tests and examples",
                },
                ("14", "109"): {
                    "title": "Resolved bugs",
                },
                ("14", "110"): {
                    "title": "Implemented suggestions",
                },
                ("14", "11"): {
                    "title": "Example page",
                },
            },
        },
        {
            "url": "https://xenforo.com/community/threads/darkness-free-xenforo-2-gaming-skin-deleted.154782/",
            "test_base_url": "https://xenforo.com/community/",
            "test_contents_hash": "0c49d593c65d6a3a57a27115a7659391d889a78d",
            "test_item_count": 53,
        },
        {
            "url": "https://xenforo.com/community/threads/steam-authentication-deleted.141309/",
            "test_base_url": "https://xenforo.com/community/",
            "test_contents_hash": "ba129ef0ceb4fdb92aee79e2b0ba1a4eabf24557",
            "test_item_count": 26,
        },
        {
            "url": "https://forums.tomshardware.com/categories/hardware.1/",
            "test_base_url": "https://forums.tomshardware.com/",
            "test_boards": {
                ("1", "9"): {
                    "title": "Graphics Cards",
                },
                ("1", "7"): {
                    "title": "Systems",
                },
                ("1", "3"): {
                    "title": "Components",
                },
                ("1", "2"): {
                    "title": "CPUs",
                },
                ("1", "5"): {
                    "title": "Motherboards",
                },
                ("1", "8"): {
                    "title": "Storage",
                },
                ("1", "4"): {
                    "title": "Overclocking",
                },
                ("1", "6"): {
                    "title": "Memory",
                },
                ("1", "10"): {
                    "title": "Displays",
                },
                ("1", "44"): {
                    "title": "Prebuilt & Enterprise",
                },
                ("1", "77"): {
                    "title": "Cases",
                },
                ("1", "74"): {
                    "title": "Cooling",
                },
                ("1", "75"): {
                    "title": "Power Supplies",
                },
                ("1", "78"): {
                    "title": "Raspberry Pi & Single Board Computers",
                },
            },
        },
        {
            "url": "https://forums.tomshardware.com/threads/windows-11.3707807/",
            "test_base_url": "https://forums.tomshardware.com/",
            "test_contents_hash": "5c6b603554ae10ad04c1d75918edeed4ec5875f5",
            "test_item_count": 96,
        },
    ]

    _board_item_css = "div.structItem--thread"
    _board_next_page_css = "a.pageNav-jump--next"
    _thread_item_css = "article.message"
    _thread_next_page_css = "a.pageNav-jump--next"

    _category_class_regex = re.compile(r"^block--category(\d+)$")
    _board_class_regex = re.compile(r"^node--id(\d+)$")
    _thread_class_regex = re.compile(r"^js-threadListItem-(\d+)$")
    _thread_key_regex = re.compile(r"^thread-(\d+)$")
    _post_id_regex = re.compile(r"^post-(\d+)$")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.try_get(
            normalize_url(url, remove_suffixes=[], append_slash=False),
            should_cache=True,
            should_retry=False,
        )

        if not re.search(r'<html[^>]+id="XF"', response.text, re.MULTILINE):
            return None

        soup = Soup(response.content)

        data_nav_id_anchor = soup.find("a", attrs={"data-nav-id": "forums"})
        base_url = normalize_url(urljoin(url, data_nav_id_anchor.get("href")))
        if not base_url:
            return None

        return XenforoExtractor(session, base_url, options)

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url, should_cache=True)
        soup = Soup(response.content)

        block_category_divs = soup.find_all("div", class_=self._category_class_regex)
        for block_category_div in block_category_divs:
            category_header = block_category_div.find("h2", class_="block-header")
            category_anchor = category_header.find("a")
            category_id = regex_match(
                self._category_class_regex, block_category_div.get_list("class")
            ).group(1)
            category_href = category_header.find("a").get("href")

            self._set_board(
                path=(category_id,),
                url=urljoin(response.url, category_href),
                origin=response.url,
                data={},
                title=category_anchor.string.strip(),
                are_subboards_fetched=True,
            )

            node_id_divs = block_category_div.find_all(
                "div", class_=self._board_class_regex
            )

            for node_id_div in node_id_divs:
                subboard_id = regex_match(
                    self._board_class_regex, node_id_div.get_list("class")
                ).group(1)

                node_description_anchor = node_id_div.find(
                    "a", attrs={"data-shortcut": "node-description"}
                )

                href = node_description_anchor.get("href")

                self._set_board(
                    path=(category_id, subboard_id),
                    url=urljoin(self.base_url, href),
                    origin=response.url,
                    data={},
                    title=node_description_anchor.string.strip(),
                )

        self._fetch_lower_boards(self.root)

    def _do_fetch_subboards(self, board: Board):
        pass

    def _get_node_from_url(self, url: str):
        response = self._session.get(url, should_cache=True)
        soup = Soup(response.content)

        if breadcrumbs_ul := soup.try_find("ul", class_="p-breadcrumbs"):
            breadcrumb_anchors = breadcrumbs_ul.find_all(
                "a", attrs={"itemprop": "item"}
            )

            if len(breadcrumb_anchors) <= 1:
                return self.root
        else:
            return self.root

        # Thread.
        if soup.try_find("article"):
            board_url = urljoin(url, breadcrumb_anchors[-2].get("href"))
            block_div = soup.find(
                "div", class_="block-container", attrs={"data-lb-id": True}
            )
            thread_id = regex_match(
                self._thread_key_regex, block_div.get("data-lb-id")
            ).group(1)
            title_h1 = soup.find("h1", class_="p-title-value")

            for cur_board in self._boards:
                if cur_board.url == board_url:
                    return Thread(
                        path=cur_board.path + (thread_id,),
                        url=urljoin(self.base_url, url),
                        origin=response.url,
                        data={},
                        title=title_h1.string,
                    )
        # Board.
        else:
            # Instead of normalizing the URL, we go the easy way: we get the OpenGraph URL.
            board_href = soup.find("meta", attrs={"property": "og:url"}).get("content")

            for cur_board in self._boards:
                if cur_board.url == board_href:
                    return cur_board

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, subboard_id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _extract_board_page_thread(
        self, board: Board, state: PageState, response: Response, tag: SoupTag
    ):
        thread_id = regex_match(
            self._thread_class_regex, tag.get_list("class")[-1]
        ).group(1)

        title_div = tag.find("div", class_="structItem-title")
        title_anchor = title_div.find("a", attrs={"data-tp-primary": True})

        url = urljoin(self.base_url, title_anchor.get("href"))

        return Thread(
            path=board.path + (thread_id,),
            url=url,
            origin=response.url,
            data={},
            title=title_anchor.string,
        )

    def _extract_thread_page_post(
        self, thread: Thread, state: PageState, response: Response, tag: SoupTag
    ):
        bbwrapper_div = tag.find("div", class_="bbWrapper")
        message_attribution_ul = tag.find("ul", class_="message-attribution-main")
        url_anchor = message_attribution_ul.find("a")
        time_tag = message_attribution_ul.find("time")

        return Post(
            path=thread.path,
            subpath=(
                regex_match(self._post_id_regex, tag.get("data-content")).group(1),
            ),
            url=urljoin(state.url, url_anchor.get("href")),
            origin=response.url,
            data={},
            author=tag.get("data-author"),
            creation_time=time_tag.get("datetime"),
            content=bbwrapper_div.string,
        )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from .common import SimulatedWriter


# The WARC writer does nothing, as both WARC recording and writing is actually performed by
# `Session`.
class WarcWriter(SimulatedWriter):
    pass
